#!/usr/bin/env python3

from __future__ import annotations

__all__=['find_all_image_paths', 'describe_images_and_cache', 'image_search', 'load_easyocr', 'easyocr', 'gpu_mem_avail', 'empty_model_cache']

import gc
import logging
import os
import subprocess
import textwrap
from collections import defaultdict
from functools import lru_cache
from pathlib import Path

import torch

from imfind import etc

from .image import image_to_text, load_model_image_to_text
from .image_and_text import (image_and_text_to_text,
                             load_model_image_and_text_to_text)

logger = logging.getLogger("imfind")
logger.setLevel(logging.INFO)
formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
handler = logging.StreamHandler()
handler.setFormatter(formatter)
logger.addHandler(handler)

line_break = '#'*80
llava_error_msg = textwrap.dedent("""
{}
Torch detected gpu, tried to use LLaVa1.5 for image-to-text but inference failed due to the following error:
{}

Using smaller but faster Salesforce/BLIP (image-captioning-large) model instead.
{}
""").strip()

def gpu_mem_avail():

    available_memory = None
    if torch.cuda.is_available():
        available_memory = 0
        for i in range(torch.cuda.device_count()):
            total_memory = torch.cuda.get_device_properties(i).total_memory
            reserved_memory = torch.cuda.memory_reserved(i)
            available_memory += total_memory - reserved_memory

    return available_memory


@lru_cache(maxsize=1)
def load_easyocr():
    import easyocr

    # this needs to run only once to load the model into memory
    reader = easyocr.Reader(etc.easyocr_languages)
    return reader


def easyocr(image):
    """
    image: bytes, numpy array or str path

    EasyOCR has some limits, so wrap this in a try except.
    - File extension support: png, jpg, tiff.
    - File size limit: 2 Mb.
    - Image dimension limit: 1500 pixel.
    - Possible Language Code Combination: Languages sharing the same written script (e.g. latin) can be used together.
      English can be used with any language.
    """
    text = ""
    try:
        reader = load_easyocr()
        text = ' '.join(reader.readtext(image, detail=0))
        if text: text = 'Text extracted from image:\n' + text
    except Exception as e:
        logger.error(f"Could not run EasyOCR due to the following error:\n{e}")

    return text


def empty_model_cache():
    #logger.info(f"GPU Memory available BEFORE clearing loaded models: {gpu_mem_avail() / 1024**3} GB")
    load_model_image_to_text.cache_clear()
    load_model_image_and_text_to_text.cache_clear()
    load_easyocr.cache_clear()
    torch.cuda.empty_cache()
    _ = gc.collect()
    #logger.info(f"GPU Memory available AFTER clearing all loaded models: {gpu_mem_avail() / 1024**3} GB")


@lru_cache(maxsize=1)
def check_image_and_text():
    # In a separate process try to load the LLava model and
    # run single inference to see if model can be used successfully.
    # Tries to use EasyOCR for both models, but doesn't fail if ocr fails for any reason.
    """
    If the model cannot be loaded into memory successfully and raises an exception mid-load,
    i.e., some layers of model are loaded onto GPU but the full model cannot be loaded,
    a CUDA OOM error is thrown but the partially loaded model is not deleted from memory.
    And since an exception is raised before the model can be loaded, we don't have access to
    the model variable to delete it and free up memory. Hence this hack.
    """

    import imfind

    if os.environ.get("IMFIND_USE_LLAVA")=="False": return False
    try:
        # if this file exists, already checked on that device, so return True
        _ = Path(etc.imfind_use_llava_path).expanduser().resolve(strict=True)
        return True

    except FileNotFoundError as e:
        logger.info(f"File {etc.imfind_use_llava_path} not found.\n\
                     Running test script to determine if llava can be loaded and used successfully.")

        test_path = os.path.join(imfind.__path__[0], 'lib/test_load_llava.py')
        try:
            _ = subprocess.run(
                ["python3", test_path],
                capture_output=True,
                check=True,
            )

            file_path = Path(etc.imfind_use_llava_path).expanduser()
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, 'w') as file:
                file.write("True")
            logger.info("Was able to successfully load and use LLaVa model.\n")
            return True

        except subprocess.CalledProcessError as e:
            logger.error(llava_error_msg.format(line_break, e.stderr, line_break), exc_info=True)

    torch.cuda.empty_cache()
    _ = gc.collect()
    return False


def find_all_image_paths(directory: Path, file_types: list[str], include_hidden=False) -> list[Path]:
    """
    Find all image files in the given directory. By default, excludes any hidden directories,
    ones starting with '.' and '__', and their contents.
    """

    def should_include(file: Path) -> bool:
        # Skip hidden directories
        if not include_hidden:
            for parent in file.parents:
                if parent.name.startswith('.') or parent.name.startswith('__'):
                    return False
        return True

    # Find all image files
    image_paths = []
    for file_type in file_types:
        image_paths.extend([file for file in directory.rglob(f'*.{file_type}') if should_include(file)])
    
    return image_paths
    

def describe_images_and_cache(images: list[Path], prompt: str, use_cache=True) -> dict[str]:
    """
        images: List of paths to image files
        prompt: prompt used to generate descriptions
        
        Note: in image_and_text_to_text the bytes cached include both the image and text bytes, so if prompt is unchanged then cache can be reused.
    """
    # maps from image abs paths to their descriptions
    descriptions = defaultdict(str)
    # if gpu is available, only then use the bigger LLaVa model. By default, use smaller BLIP model
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    
    use_llava_success = False
    if device != 'cpu': use_llava_success = check_image_and_text()

    for img_path in images:
        k = str(img_path)
        try:
            if device != 'cpu' and use_llava_success:
                try:
                    descriptions[k] = image_and_text_to_text(img_path, prompt, use_cache)
                except Exception as e:
                    logger.error(llava_error_msg.format(line_break, e, line_break), exc_info=True)
                    use_llava_success = False
                    empty_model_cache()

                    descriptions[k] = image_to_text(img_path, use_cache)
            else:
                descriptions[k] = image_to_text(img_path, use_cache)
        except Exception as e:
            descriptions[k] = img_path.name
            logger.error(f"Could not describe image '{k}' due to the following error:\n{e}\nUsing file name for description instead.\n", exc_info=True)

    # makes space for embedding model and any downstream tasks (in cases of limited gpu memory)
    empty_model_cache()
    return descriptions


def image_search(user_img_desc: str, gen_desc_prompt: str, directory: Path, file_types: list[str], use_cache=True, include_hidden=False, embed_size='large') -> list[str]:

    from embd import EmbedFlag, Space
    
    images = find_all_image_paths(directory, file_types, include_hidden)
    descriptions = describe_images_and_cache(images, gen_desc_prompt, use_cache)
    space = Space(EmbedFlag(embed_size))
    
    user_desc_embed = space.think(user_img_desc)

    # for each image_path:desc pair, embed the descriptions and dot product them with the desired image description embedding to get similarity score for each image
    
    desc_embeds_scores = {path: user_desc_embed @ space.think(desc).T for path, desc in descriptions.items()}
    sorted_path_scores = sorted(desc_embeds_scores.items(), key=lambda item: item[1], reverse=True)
    
    # return img paths (like find command does) sorted by most to least familiar
    return [pair[0] for pair in sorted_path_scores]




#!/usr/bin/env python3

__all__ = ['main']

import argparse
import os
import sys
import platform
from pathlib import Path

from imfind import etc


def parse_args(argv):

    def check_nonneg(value):
        ivalue = int(value)
        if ivalue < 0:
            raise argparse.ArgumentTypeError(f"Threshold needs to be an integer greater than or equal to 1 (n>=1).\n Passed {value}.")
        return ivalue

    parser = argparse.ArgumentParser('imfind')
    parser.add_argument('description', type=str, help="Description of the image to find.")
    parser.add_argument('-d', '--directory', type=str, help="Directory to search in. Defaults to $HOME.")
    parser.add_argument('-n', '--threshold', type=check_nonneg, default=0, help="Top n (>=1) results to output. By default lists all images found (like the find command) sorted from most relevant/similar to least.")
    parser.add_argument('-p', '--prompt', type=str, help="Additional user prompt that gets appended to default prompt. Used to generate description of images.")
    parser.add_argument('-t', '--types', type=str, nargs='*', default=etc.file_types, help=f"Additional image types to search. By default searches for images with  extensions {etc.file_types}")
    parser.add_argument('--no-cache', action='store_true', help="Do not read from existing cache. Overwrites cache with new model generations. (Internally sets and uses USE_MMRY_CACHE environment variable).")
    parser.add_argument('--include-hidden', action='store_true', help="Include hidden directories like ones starting with '.' or '__'. For example, within .cache or __pycache__ etc., Excludes these by default.")
    args = parser.parse_args(argv)
    return args

def main(argv=None):

    if argv is None:
        argv = sys.argv[1:]

    args = parse_args(argv)

    description = args.description.strip()
    if not description:
        raise Exception("Error: Empty description. Please provided a detailed description of the image to find.")

    directory = ''
    try:
        if not args.directory:
            HOME = "HOME"
            if platform.system() == "Windows": HOME = "USERPROFILE"
            directory = Path(os.environ[HOME]).expanduser().resolve(strict=True) # make it home directory
        else:
            directory = Path(args.directory.strip()).expanduser().resolve(strict=True)
    except Exception as e:
        raise e
    
    file_types = etc.file_types + (args.types or [])
    file_types += [e.capitalize() for e in file_types]
    file_types = list(set(file_types))

    thres = args.threshold
    final_prompt = etc.default_prompt + (args.prompt or '').strip()

    use_cache = False if args.no_cache else True
    include_hidden = bool(args.include_hidden)

    # do the image search
    from imfind import image_search

    top = image_search(user_img_desc=description, gen_desc_prompt=final_prompt, 
                       directory=directory, file_types=file_types, use_cache=use_cache, include_hidden=include_hidden)

    if thres:
        print('\n'.join(top[:thres]))
    else:
        print('\n'.join(top))

if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python3

### Adapted from notarealdeveloper/kern repository 
### https://github.com/notarealdeveloper/kern/blob/master/src/kern/image/lib/image_and_text.py 

__all__ = [
    'image_and_text_to_text', 'load_model_image_and_text_to_text'
]

import gc
import io
import os
import re
from functools import lru_cache

import torch

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

def image_and_text_to_text(image, text, use_cache=True):

    import assure
    from mmry import Cache

    from imfind import easyocr

    image_bytes = assure.bytes(image)
    text_bytes = text.encode()
    bytes = image_bytes + text_bytes

    cache = Cache('image_and_text_to_text')
    if use_cache and cache.have_blob(bytes):
        return cache.load_blob(bytes).decode()

    text = image_and_text_to_text_nocache(image_bytes, text)

    ocr_text = easyocr(image_bytes)
    text += '\n' + ocr_text + '\n'
    cache.save_blob(bytes, text)
    return text

def image_and_text_to_text_nocache(image_bytes, text):

    import PIL.Image
    from pillow_heif import register_heif_opener
    register_heif_opener()

    file = io.BytesIO(image_bytes)
    image = PIL.Image.open(file)

    # Define a chat histiry and use `apply_chat_template` to get correctly formatted prompt
    # Each value in "content" has to be a list of dicts with types ("text", "image")
    conversation = [
        {

        "role": "user",
        "content": [
            {"type": "text", "text": text},
            {"type": "image"},
            ],
        },
    ]

    processor, model = load_model_image_and_text_to_text()

    prompt = processor.apply_chat_template(conversation, add_generation_prompt=True)
    encoding = processor(images=image, text=prompt, return_tensors='pt').to(device, torch.float16)

    generate_ids = model.generate(**encoding, max_new_tokens=1024, do_sample=False)
    output = processor.decode(generate_ids[0], skip_special_tokens=True)

    del encoding # frees up any gpu memory taken by this
    torch.cuda.empty_cache()
    _ = gc.collect()

    m = re.search("ASSISTANT:", output)
    return output[m.end():].strip()

@lru_cache(maxsize=1)
def load_model_image_and_text_to_text():
    from transformers import AutoProcessor, LlavaForConditionalGeneration
    name = "llava-hf/llava-1.5-7b-hf"
    processor = AutoProcessor.from_pretrained(name)

    # Using device_map='auto' with single GPU tends to slow it down significantly (by ~4 times)
    if torch.cuda.is_available() and torch.cuda.device_count()>1:
        model = LlavaForConditionalGeneration.from_pretrained(name, torch_dtype=torch.float16, low_cpu_mem_usage=True, device_map='auto')
    else:
        model = LlavaForConditionalGeneration.from_pretrained(name, torch_dtype=torch.float16, low_cpu_mem_usage=True,).to(device)
    return (processor, model)



from pathlib import Path

import assure

from imfind import (easyocr, etc, image_and_text_to_text, image_search,
                    image_to_text)


def test_image_search():
    # CPU/GPU usage determined within
    path = Path("./include/examples/").expanduser().resolve()
    user_description = "raccooon and cars" # embedding model can handle typo's too
    img_paths = image_search(user_description, etc.default_prompt, path, etc.file_types)
    top_two_paths = ' '.join(img_paths[:3])
    assert "racoon" in top_two_paths
    assert "misc-3" in top_two_paths


def test_image_to_text():
    text = image_to_text("./include/examples/animal-pics/puppy-pic-1.jpeg", use_cache=False)
    text2 = image_to_text("./include/examples/misc/food-1.png", use_cache=False)
    assert "puppies" in text
    assert "pancakes" in text2


def example_image_and_text_to_text():
    """ Example usage of LLaVa model which also takes additional generation prompt as input.
        Only recommended to run on GPU, extremely slow on CPU due to large model """
    
    text = image_and_text_to_text("./include/examples/animal-pics/puppy-pic-1.jpeg", etc.default_prompt, use_cache=False)
    print(text)

def test_easyocr():
    # read with image bytes
    image1 = Path("./include/examples/texts/text-1.jpeg").expanduser().resolve()
    image_bytes1 = assure.bytes(image1)
    ocr_text1 = easyocr(image_bytes1)

    # read with image path as str
    image2 = "./include/examples/texts/text-2.jpeg"
    ocr_text2 = easyocr(image2)
    
    assert "Privet Drive" in ocr_text1
    assert "Half Blood Prince" in ocr_text2




#!/usr/bin/env python3

### Adapted from notarealdeveloper/kern repository 
### https://github.com/notarealdeveloper/kern/blob/master/src/kern/image/lib/image.py 

__all__ = [
    'image_to_text', 'load_model_image_to_text',
]

import io
import os
from functools import lru_cache

import torch

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

def image_to_text(image, use_cache=True):

    import assure
    from mmry import Cache

    from imfind import easyocr

    bytes = assure.bytes(image)

    cache = Cache('image_to_text')
    if use_cache and cache.have_blob(bytes):
        return cache.load_blob(bytes).decode()

    text = image_to_text_nocache(bytes)

    ocr_text = easyocr(bytes)
    text += '\n' + ocr_text + '\n'
    cache.save_blob(bytes, text.encode())
    return text


def image_to_text_nocache(bytes):

    import PIL.Image
    from pillow_heif import register_heif_opener
    register_heif_opener()

    file = io.BytesIO(bytes)
    image = PIL.Image.open(file).convert('RGB')

    processor, model = load_model_image_to_text()
    
    encoding = processor(image, return_tensors="pt").to(device)

    generate_ids = model.generate(**encoding, max_new_tokens=1024, temperature=1.0)
    output =  processor.decode(generate_ids[0], skip_special_tokens=True)
    return output


@lru_cache(maxsize=1)
def load_model_image_to_text():
    from transformers import BlipForConditionalGeneration, BlipProcessor
    name = "Salesforce/blip-image-captioning-large"
    processor = BlipProcessor.from_pretrained(name)
    model = BlipForConditionalGeneration.from_pretrained(name).to(device)
    return (processor, model)



default_prompt = """
What is shown in this image?
Generate a detailed description of this image.
Include any specific details on background colors, patterns, themes, settings/context (for example if it's a search page results, texting platform screenshot, pic of scenery etc.,), what might be going in in the picture (activities, conversations), what all and how many objects, animals and people are present, their orientations and activities, etc.,
Besides a general description, include any details that might help uniquely identify the image.
"""

file_types = ['jpg', 'jpeg', 'png', 'gif', 'bmp', 'webp', 'tiff', 'heic']

# See here for full available list https://www.jaided.ai/easyocr/
easyocr_languages = ['en', 'ch_tra']

imfind_use_llava_path = '~/.cache/imfind/use_llava.txt'



#!/usr/bin/env python3

import os

import requests

from imfind import image_and_text_to_text
from imfind.etc import default_prompt

# Try a test run of loading big model (LLaVa 1.5) into memory.
# If success please set IMFIND_USE_LLAVA to "True"
 
url = "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/transformers/tasks/ai2d-demo.jpg"
image = requests.get(url, stream=True).raw
description = image_and_text_to_text(image, default_prompt, use_cache=False)
